/* Copyright (C) 1992-1998 The Geometry Center
 * Copyright (C) 1998-2000 Stuart Levy, Tamara Munzner, Mark Phillips
 *
 * This file is part of Geomview.
 * 
 * Geomview is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 * 
 * Geomview is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with Geomview; see the file COPYING.  If not, write
 * to the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139,
 * USA, or visit http://www.gnu.org.
 */

#if HAVE_CONFIG_H
# include "config.h"
#endif

#if 0
static char copyright[] = "Copyright (C) 1992-1998 The Geometry Center\n\
Copyright (C) 1998-2000 Stuart Levy, Tamara Munzner, Mark Phillips";
#endif


/* Authors: Charlie Gunn, Stuart Levy, Tamara Munzner, Mark Phillips */

#include "bboxP.h"

BBox *BBoxMinMax(BBox *bbox, HPoint3 *min, HPoint3 *max)
{
  static HPoint3 nullpoint = {0,0,0,1};

  if (bbox != NULL) {
    HPtNToHPt3(bbox->min, NULL, min);
    HPtNToHPt3(bbox->max, NULL, max);
  } else {
    *min = nullpoint;
    *max = nullpoint;
  }
  return bbox;
}

BBox *BBoxMinMaxND(BBox *bbox, HPointN **min, HPointN **max)
{
  if (bbox != NULL) {
    *min = HPtNCopy(bbox->min, *min);
    *max = HPtNCopy(bbox->max, *max);
  } else {
    *min = NULL;
    *max = NULL;
  }
  return bbox;
}

/*
 * Local Variables: ***
 * c-basic-offset: 2 ***
 * End: ***
 */
